/*
 *  Copyright (C) 2005-2006 mark307
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <windowsx.h>
#include <tchar.h>
#include <stdlib.h>

#define KMELEON_PLUGIN_EXPORTS
#include <kmeleon_plugin.h>
#include <KMeleonConst.h>
#include <resource.h>

extern "C" {
#include <lua.h>
#include <lualib.h>
#include <lauxlib.h>
}

#include "luamacros.h"
#include "luaut.h"
#include "macroset.h"
#include "kmapi.h"
#include "kmenv.h"
#include "w32.h"

#define RAISE_ERROR return luaL_error(L, "internal error");

extern kmeleonPlugin kPlugin;
extern kmeleonFunctions *kFuncs;
extern MacroSet* g_lpMacroset;


/**
   Lua macro Plugin for K-Meleon
   
   * K-Meleon functions
   
   ** Defined variables
   
   PREFPATH
   
      A path of current preferences.
   
   LIBPATH
   
      A path of addtional Lua library.
      Preference "kmeleon.plugins.luamacros.path" is copyed to this.
      If above preference is none, this value is same as PREFPATH.
   
   BROWSERPATH
   
      A path of folder that has k-meleon.exe.
   
   KMEL_PLUGIN_VER
   
      Version of plugin API. 513 is 0.9, 514 is 1.0.
   
   TYPE_BOOL TYPE_INT TYPE_STRING
   TYPE_UNISTRING TYPE_TSTRING (1.0 feature)
   
      Type identifier just like BOOL, INT, STRING.
   
   OPEN_NORMAL OPEN_NEW OPEN_BACKGROUND
   OPEN_NEWTAB OPEN_BACKGROUNDTAB OPEN_CLONE (1.0 feature)
   
      State of new open window, used on argument of NavigateTo.
   
   STORE_ANYWHERE STORE_IN_MEMORY STORE_ON_DISK STORE_ON_DISK_AS_FILE
   
      Cache control
   
   UWM_UPDATESESSIONHISTORY UWM_REFRESHTOOLBARITEM UWM_UPDATEBUSYSTATE
   UWM_REFRESHMRULIST UWM_NEWWINDOW UWM_PERSIST_SET UWM_PERSIST_GET
   UWM_PERSIST_SHOW
   
   
   ** Event functions
   
   OnInit OnSetup OnStartup OnOpenWindow OnLoad OnFindNick
   OnCloseWindow OnCloseGroup OnDestroy OnQuit OnExit
   OnActivateWindow OnCloseTab OnCreateTab OnSwitchTab (1.0 feature)
   
   ** Defined functions
   
 */

const lut_IntReg consts[] = {
	/*
	 * types
	 */
	{"TYPE_BOOL", PREF_BOOL},
	{"TYPE_INT", PREF_INT},
	{"TYPE_STRING", PREF_STRING},
#if (KMEL_PLUGIN_VER) > 0x0201 /* >= 1.0 */
	{"TYPE_UNISTRING", PREF_UNISTRING},
	{"TYPE_TSTRING", PREF_TSTRING},
#endif
	/* 
	 * NavigateTo
	 */
	{"OPEN_NORMAL", OPEN_NORMAL},
	{"OPEN_NEW", OPEN_NEW},
	{"OPEN_BACKGROUND", OPEN_BACKGROUND},
#if (KMEL_PLUGIN_VER) == 0x0202 /* == 1.0 */
	{"OPEN_NEWTAB", OPEN_NEWTAB},
	{"OPEN_BACKGROUNDTAB", OPEN_BACKGROUNDTAB},
#endif
#if (KMEL_PLUGIN_VER) > 0x0201 /* >= 1.0 */
	{"OPEN_CLONE", OPEN_CLONE},
#endif
	/*
	 * from nsICache.h
	 */
	{"STORE_ANYWHERE", 0},
	{"STORE_IN_MEMORY", 1},
	{"STORE_ON_DISK", 2},
	{"STORE_ON_DISK_AS_FILE", 3},
	/*
	 * from KMeleonConst.h
	 */
	{"UWM_UPDATESESSIONHISTORY", UWM_UPDATESESSIONHISTORY},
	{"UWM_REFRESHTOOLBARITEM", UWM_REFRESHTOOLBARITEM},
	{"UWM_UPDATEBUSYSTATE", UWM_UPDATEBUSYSTATE},
	{"UWM_REFRESHMRULIST", UWM_REFRESHMRULIST},
	{"UWM_NEWWINDOW", UWM_NEWWINDOW},
	{"UWM_PERSIST_SET", UWM_PERSIST_SET},
	{"UWM_PERSIST_GET", UWM_PERSIST_GET},
	{"UWM_PERSIST_SHOW", UWM_PERSIST_SHOW},
	/*
	 * plugin interface version
	 */
	{"KMEL_PLUGIN_VER", KMEL_PLUGIN_VER},
#if (KMEL_PLUGIN_VER) > 0x0202 /* >= 1.1 */
	/* FolderType */
	{"RootFolder", RootFolder},
	{"DefSettingsFolder", DefSettingsFolder},
	{"UserSettingsFolder", UserSettingsFolder},
	{"ProfileFolder", ProfileFolder},
	{"PluginsFolder", PluginsFolder},
	{"UserPluginsFolder", UserPluginsFolder},
	{"SkinsFolder", SkinsFolder},
	{"UserSkinsFolder", UserSkinsFolder},
	{"ResFolder", ResFolder},
	{"CurrentSkinFolder", CurrentSkinFolder},
	/* WindowVarType */
	{"Window_UrlBar", Window_UrlBar},
	{"Window_Charset", Window_Charset},
	{"Window_Title", Window_Title},
	{"Window_TextZoom", Window_TextZoom},
	{"Window_URL", Window_URL},
	/* Read Only WindowVarType */
	{"Window_SelectedText", Window_SelectedText},
	{"Window_LinkURL", Window_LinkURL},
	{"Window_ImageURL", Window_ImageURL},
	{"Window_FrameURL", Window_FrameURL},
	/* MENUTYPE */
	{"MENU_COMMAND", MENU_COMMAND},
	{"MENU_POPUP", MENU_POPUP},
	{"MENU_INLINE", MENU_INLINE},
	{"MENU_PLUGIN", MENU_PLUGIN},
	{"MENU_SEPARATOR", MENU_SEPARATOR},
#endif
	/* end */
	{NULL, 0},
};


static int get_commandid(lua_State *L, int idx)
{
	int id = -1;
	char *cmd_id;
	
	if (lua_isnumber(L, idx)) {
		id = lua_tointeger(L, idx);
	} else if (lua_isstring(L, idx)) {
		cmd_id = lut_tostring(L, idx);
		try {
			id = kFuncs->GetID(cmd_id);
		} catch (...) {
			RAISE_ERROR;
		}
		lua_pop(L, 1); /* delete cmd_id */
	} else {
		return luaL_typerror(L, idx, "number or string");
	}
	return id;
}


/*
   Callback
*/
LUALIB_API int L_DoError(lua_State *L)
{
	HWND hwnd;
	
	hwnd = get_hwnd(L);
	MessageBox(hwnd, luaL_optstring(L, -1, "Unknown error"),
	           kPlugin.dllname, MB_ICONINFORMATION | MB_OK);
	return 0;
}

LUALIB_API int L_DoExit(lua_State *L)
{
	lua_pushstring(L, "terminated by os.exit()");
	return lua_error(L);
}

LUALIB_API int L_DoPrint(lua_State *L)
{
	int n = lua_gettop(L);
	const char *s;
	
	lua_concat(L, n);
	s = lua_tostring(L, -1);
	if (s) {
		// kFuncs->SetStatusBarText(s);
		OutputDebugString(s);
	}
	return 0;
}

LUALIB_API void L_CountHook(lua_State *L, lua_Debug *D)
{
	HWND hwnd;
	int result;
	
	hwnd = get_hwnd(L);
	result = MessageBox(hwnd, "continue?", PLUGIN_NAME,
	                    MB_ICONQUESTION | MB_YESNO);
	if (result == IDNO) {
		lua_pushstring(L, "terminated by user");
		lua_error(L);
	}
}

/**
   plugin(plugin, command)
   
      Run a plugin command.
   
   plugin  : string : plugin name
   command : string : plugin command
 */
int km_plugin(lua_State *L)
{
	int cmd = 0;
	HWND hwnd;
	const char *plugin;
	char *param;
	
	luaL_checkany(L, 2);
	hwnd = (HWND) get_hwnd(L);
	plugin = luaL_checklstring(L, 1, NULL);
	param = lut_checkstring(L, 2);
	try {
		kFuncs->SendMessage(plugin, kPlugin.dllname, "DoAccel",
		                    (long) param, (long) &cmd);
	} catch (...) {
		RAISE_ERROR;
	}
	if (cmd > 0) {
		SendMessage(hwnd, WM_COMMAND, cmd, (LPARAM) NULL);
	}
	return 0;
}


/**
   pluginmsg(plugin, message [, param1 [, param2]])
   
      Run a plugin command.
   
   plugin         : string : plugin name
   message        : string : message passed to plugin
   param1, param2 : string : parameter passed to plugin
 */
int km_pluginmsg(lua_State *L)
{
	const char *plugin, *message;
	char *param1, *param2;
	
	luaL_checkany(L, 2);
	lua_settop(L, 4); // for optstring
	plugin  = luaL_checklstring(L, 1, NULL);
	message = luaL_checklstring(L, 2, NULL);
	param1 = lut_optstring(L, 3, "");
	param2 = lut_optstring(L, 4, "");
	
	try {
		kFuncs->SendMessage(plugin, kPlugin.dllname, message,
		                    (long) param1, (long) param2);
	} catch (...) {
		RAISE_ERROR;
	}
	return 0;
}


/**
   pluginmsgex(plugin, message, [param,] type)
   
      Run a plugin command, and get return value.
   
   plugin  : string : plugin name
   message : string : message passed to plugin
   param   : string : parameter passed to plugin
   type    : TYPE_BOOL | TYPE_INT | TYPE_STRING : type of return value
   
   return : boolean | number | string
*/
int km_pluginmsgex(lua_State *L)
{
	int n = lua_gettop(L);
	const char *plugin, *message;
	char *param = NULL;
	PREFTYPE type;
	char *cRetval = NULL;
	int nRetval = 0;
	
	luaL_checkany(L, 3);
	plugin  = luaL_checklstring(L, 1, NULL);
	message = luaL_checklstring(L, 2, NULL);
	if (n > 3) {
		param = lut_checkstring(L, 3);
		type = (PREFTYPE) luaL_checkinteger(L, 4);
	} else {
		type = (PREFTYPE) luaL_checkinteger(L, 3);
	}
	
	switch (type) {
	case PREF_BOOL:
		try {
			kFuncs->SendMessage(plugin, kPlugin.dllname, message,
			                    (long) param, (long) &nRetval);
		} catch (...) {
			RAISE_ERROR;
		}
		lua_pushboolean(L, nRetval);
		break;
	case PREF_INT:
		try {
			kFuncs->SendMessage(plugin, kPlugin.dllname, message,
			                    (long) param, (long) &nRetval);
		} catch (...) {
			RAISE_ERROR;
		}
		lua_pushinteger(L, (int) nRetval);
		break;
	case PREF_STRING:
		try {
			kFuncs->SendMessage(plugin, kPlugin.dllname, message,
			                    (long) param, (long) &cRetval);
		} catch (...) {
			free(cRetval);
			RAISE_ERROR;
		}
		lua_pushstring(L, cRetval);
		free(cRetval);
		break;
	default:
		return luaL_argerror(L, (n > 3) ? 4 : 3, "invalid pref type");
	}
	return 1;
}


/**
   GetCommandIDs([n])
   
      Wrapper of GetCommandIDs K-Meleon API.
   
   n : number : necessary number of ID
   
   return : number : first ID
 */
int km_GetCommandIDs(lua_State *L)
{
	int id, nid;
	
	luaL_checkany(L, 1);
	nid = luaL_optinteger(L, 1, 1);
	try {
		id = kFuncs->GetCommandIDs(nid);
	} catch (...) {
		RAISE_ERROR;
	}
	lua_pushinteger(L, id);
	return 1;
}


/**
   NavigateTo(url [, windowstate])
   
      Wrapper of NavigateTo K-Meleon API.
      Open a page.
   
   url : string : URL
   windowstate : OPEN_NORMAL |
                 OPEN_NEW |
                 OPEN_BACKGROUND : specify open method
 */
int km_NavigateTo(lua_State *L)
{
	const char *url;
	int windowstate;
	HWND hwnd;
	
	luaL_checkany(L, 1);
	url = luaL_checklstring(L, 1, NULL);
	windowstate = luaL_optinteger(L, 2, (lua_Integer) OPEN_NORMAL);
	hwnd = get_hwnd(L);
	
	if (windowstate != OPEN_NORMAL &&
	    windowstate != OPEN_NEW &&
	    windowstate != OPEN_BACKGROUND) {
		return luaL_argerror(L, 2, "invalid open type");
	}
	try {
		kFuncs->NavigateTo(url, windowstate, hwnd);
	} catch (...) {
		RAISE_ERROR;
	}
	return 0;
}


/**
   GetDocInfo()
  
     Wrapper of GetDocInfo K-Meleon API.
     Gets the title and URL of current page.
   
   return : string, string : First value is title of current page,
                             second value is url of current page.
 */
int km_GetDocInfo(lua_State *L)
{
	kmeleonDocInfo *docinfo;
	HWND hwnd;
	
	hwnd = get_hwnd(L);
	try {
		docinfo = kFuncs->GetDocInfo(hwnd);
	} catch (...) {
		RAISE_ERROR;
	}
	if (docinfo) {
		lua_pushstring(L, docinfo->title);
		lua_pushstring(L, docinfo->url);
		return 2;
	}
	return 0;
}


/**
   getpref(type, preference)
   
      Gets a user preference.
   
   type : TYPE_BOOL | TYPE_INT | TYPE_STRING : specify preference type
   preference : string : name of preference.
   
   return : boolean | number | string
 */
#if (KMEL_PLUGIN_VER) > 0x0201
int km_getpref(lua_State *L)
{
	PREFTYPE type;
	const char *pref;
	void *cRetval = NULL;
	long len;
	int nRetval = 0;
	
	luaL_checkany(L, 2);
	type = (PREFTYPE) luaL_checkinteger(L, 1);
	pref = luaL_checkstring(L, 2);

	switch (type) {
	case PREF_BOOL:
		try {
			kFuncs->GetPreference(type, pref, &nRetval, &nRetval);
		} catch (...) {
			RAISE_ERROR;
		}
		lua_pushboolean(L, nRetval);
		break;
	case PREF_INT:
		try {
			kFuncs->GetPreference(type, pref, &nRetval, &nRetval);
		} catch (...) {
			RAISE_ERROR;
		}
		lua_pushinteger(L, nRetval);
		break;
	case PREF_STRING:
		try {
			len = kFuncs->GetPreference(type, pref, NULL, (void *) "");
		} catch (...) {
			RAISE_ERROR;
		}
		cRetval = lua_newuserdata(L, len + 1);
		try {
			kFuncs->GetPreference(type, pref, cRetval, (void *) "");
		} catch (...) {
			RAISE_ERROR;
		}
		lua_pushlstring(L, (char *) cRetval, len);
		break;
	case PREF_UNISTRING:
		try {
			len = kFuncs->GetPreference(type, pref, 0, (void *) L"") + 1;
		} catch (...) {
			RAISE_ERROR;
		}
		cRetval = lua_newuserdata(L, sizeof(wchar_t) * len);
		try {
			kFuncs->GetPreference(type, pref,cRetval, (void *) L"");
		} catch (...) {
			RAISE_ERROR;
		}
		lua_pushlstring(L, (char *) cRetval, sizeof(wchar_t) * len);
		break;
	default:
		return luaL_argerror(L, 1, "invalid pref type");
	}
	return 1;
}
#else
int km_getpref(lua_State *L)
{
	PREFTYPE type;
	char *pref;
	char cRetval[4096] = "";
	int nRetval = 0;
	
	luaL_checkany(L, 2);
	type = (PREFTYPE) luaL_checkinteger(L, 1);
	pref = lut_checkstring(L, 2);

	switch (type) {
	case PREF_BOOL:
		try {
			kFuncs->GetPreference(type, pref, &nRetval, &nRetval);
		} catch (...) {
			RAISE_ERROR;
		}
		lua_pushboolean(L, nRetval);
		break;
	case PREF_INT:
		try {
			kFuncs->GetPreference(type, pref, &nRetval, &nRetval);
		} catch (...) {
			RAISE_ERROR;
		}
		lua_pushinteger(L, nRetval);
		break;
	case PREF_STRING:
		try {
			kFuncs->GetPreference(type, pref, cRetval, NULL);
		} catch (...) {
			RAISE_ERROR;
		}
		lua_pushstring(L, cRetval);
		break;
	default:
		return luaL_argerror(L, 1, "invalid pref type");
	}
	return 1;
}
#endif

/**
   setpref(type, preference, value)
   
      Sets user preferences.
   
   type : TYPE_BOOL | TYPE_INT | TYPE_STRING : specify preference type
   preference : string : name of preference.
   value : boolean | number | string
 */
int km_setpref(lua_State *L)
{
	PREFTYPE type;
	char *pref, *cValue;
	int nValue = 0;
	
	luaL_checkany(L, 3);
	type = (PREFTYPE) luaL_checkinteger(L, 1);
	pref = lut_checkstring(L, 2);
	switch (type) {
	case PREF_BOOL:
		luaL_checktype(L, 3, LUA_TBOOLEAN);
		nValue = lua_toboolean(L, 3);
		try {
			kFuncs->SetPreference(type, pref, &nValue, TRUE);
		} catch (...) {
			RAISE_ERROR;
		}
		break;
	case PREF_INT:
		nValue = luaL_checkinteger(L, 3);
		try {
			kFuncs->SetPreference(type, pref, &nValue, TRUE);
		} catch (...) {
			RAISE_ERROR;
		}
		break;
	case PREF_STRING:
		cValue = lut_checkstring(L, 3);
		try {
			kFuncs->SetPreference(type, pref, cValue, TRUE);
		} catch (...) {
			RAISE_ERROR;
		}
		break;
	default:
		return luaL_argerror(L, 1, "invalid pref type");
	}
	return 0;
}


/**
   delpref(preference)
   
      Deletes user preferences.
   
   preference : string
 */
int km_delpref(lua_State *L)
{
	char *pref;
	
	luaL_checkany(L, 1);
	pref = lut_checkstring(L, 1);
	try {
		kFuncs->DelPreference(pref);
	} catch (...) {
		RAISE_ERROR;
	}
	return 0;
}


/**
   SetStatusBarText(text)
   
      Wrapper of SetStatusBarText K-Meleon API.
      Show the message text on the status bar.
   
   text : string
 */
int km_SetStatusBarText(lua_State *L)
{
	luaL_checkany(L, 1);
	try {
		kFuncs->SetStatusBarText(luaL_checklstring(L, 1, NULL));
	} catch (...) {
		RAISE_ERROR;
	}
	return 0;
}


/**
   GotoHistoryIndex(index)
   
      Wrapper of GotoHistoryIndex K-Meleon API.
   
   index : number
 */
int km_GotoHistoryIndex(lua_State *L)
{
	try {
		kFuncs->GotoHistoryIndex(luaL_checkinteger(L, 1));
	} catch (...) {
		RAISE_ERROR;
	}
	return 0;
}


/**
   GetID(id)
   
      Wrapper of GetID K-Meleon API.
      Convert ID string like "ID_OPEN_LINK" to ID number.
   
   id : string : Name of ID.
   
   return : number : ID
 */
int km_GetID(lua_State *L)
{
	char *strID;
	int id;
	
	luaL_checkany(L, 1);
	strID = lut_checkstring(L, 1);
	try {
		id = kFuncs->GetID(strID);
	} catch (...) {
		RAISE_ERROR;
	}
	lua_pushinteger(L, id);
	return 1;
}


/**
   GetGlobalVar(type, preference)
   
      Wrapper of GetGlobalVar K-Meleon API.
   
   type : TYPE_BOOL | TYPE_INT | TYPE_STRING : specify type
   preference : string :
   
   Below preferences can accept.
      TYPE_STRING : URL LinkURL ImageURL FrameURL TITLE
                    CHARSET TextZoom SelectedText URLBAR (1.0 feature)
   
   return : boolean | number | string
 */
int km_GetGlobalVar(lua_State *L)
{
	PREFTYPE type;
	char *pref;
	char *cRetval = NULL;
	int nRetval = 0;
	
	luaL_checkany(L, 2);
	type = (PREFTYPE) luaL_checkinteger(L, 1);
	pref = lut_checkstring(L, 2);

	switch (type) {
	case PREF_BOOL:
		try {
			kFuncs->GetGlobalVar(PREF_BOOL, pref, &nRetval);
		} catch (...) {
			RAISE_ERROR;
		}
		lua_pushboolean(L, nRetval);
		break;
	case PREF_INT:
		try {
			kFuncs->GetGlobalVar(PREF_INT, pref, &nRetval);
		} catch (...) {
			RAISE_ERROR;
		}
		lua_pushinteger(L, nRetval);
		break;
	case PREF_STRING:
		try {
			nRetval = (int) kFuncs->GetGlobalVar(PREF_STRING, pref, NULL);
		} catch (...) {
			RAISE_ERROR;
		}
		if (nRetval) {
			cRetval = (char *) lua_newuserdata(L, sizeof(char) * (nRetval + 1));
			cRetval[0] = '\0';
			try {
				kFuncs->GetGlobalVar(PREF_STRING, pref, cRetval);
			} catch (...) {
				RAISE_ERROR;
			}
			lua_pushstring(L, (char *) cRetval);
		} else {
			lua_pushstring(L, "");
		}
		break;
	default:
		return luaL_argerror(L, 1, "invalid pref type");
	}
	return 1;
}


/**
   EncodeUTF8(text)
   
      Wrapper of EncodeUTF8 K-Meleon API.
      Convert native (ascii) string to UTF8
   
   text : string
   
   return : string
 */
int km_EncodeUTF8(lua_State *L)
{
	const char *astr = NULL;
	char *ustr = NULL;
	
	astr = luaL_optstring(L, 1, "");
	try {
		ustr = kFuncs->EncodeUTF8(astr);
	} catch (...) {
		RAISE_ERROR;
	}
	lua_pushstring(L, ustr);
	free(ustr);
	return 1;
}


/**
   DecodeUTF8(text)
   
      Wrapper of DecodeUTF8 K-Meleon API.
      Convert UTF8 string to native (ascii)
   
   text : string
   
   return : string
 */
int km_DecodeUTF8(lua_State *L)
{
	const char *ustr = NULL;
	char *astr = NULL;
	
	ustr = luaL_optstring(L, 1, "");
	try {
		astr = kFuncs->DecodeUTF8(ustr);
	} catch (...) {
		RAISE_ERROR;
	}
	lua_pushstring(L, astr);
	free(astr);
	return 1;
}


/**
   GetMenu(name)
   
      Wrapper of GetMenu K-Meleon API.
      Gets menu handle (HMENU) of specified name.
   
   name : string
   
   return : number (HMENU) : menu handle
 */
int km_GetMenu(lua_State *L)
{
	HMENU hmenu;
	char *menu;
	
	luaL_checkany(L, 1);
	menu = lut_checkstring(L, 1);
	try {
		hmenu = kFuncs->GetMenu(menu);
	} catch (...) {
		RAISE_ERROR;
	}
	lua_pushinteger(L, (lua_Integer) hmenu);
	return 1;
}


/**
   SetForceCharset(charset)
   
      Wrapper of SetForceCharset K-Meleon API.
      Force charactor set of current page to specified.
   
   charset : string
 */
int km_SetForceCharset(lua_State *L)
{
	char *charset;
	
	charset = lut_checkstring(L, 1);
	try {
		kFuncs->SetForceCharset(charset);
	} catch (...) {
		RAISE_ERROR;
	}
	return 0;
}


/**
   SetCheck(id, mark)
   
      Wrapper of SetCheck K-Meleon API.
      Sets checkmark to manu item.
   
   id   : number | string : menu id
   mark : boolean         : check or uncheck
 */
int km_SetCheck(lua_State *L)
{
	int id = -1, mark;
	
	luaL_checkany(L, 2);
	id = get_commandid(L, 1);
	luaL_checktype(L, 2, LUA_TBOOLEAN);
	mark = lua_toboolean(L, 2);
	if (id > 0) {
		try {
			kFuncs->SetCheck(id, mark);
		} catch (...) {
			RAISE_ERROR;
		}
	}
	return 0;
}


/**
   ClearCache(cache)
   
      Wrapper of ClearCache K-Meleon API.
   
   cache : number : STORE_ANYWHERE | STORE_IN_MEMORY |
                    STORE_ON_DISK | STORE_ON_DISK_AS_FILE
 */
int km_ClearCache(lua_State *L)
{
	int cache;
	luaL_checkany(L, 1);
	cache = luaL_checkinteger(L, 1);
	try {
		kFuncs->ClearCache(cache);
	} catch (...) {
		RAISE_ERROR;
	}
	return 0;
}


/**
   BroadcastMessage(msg, wparam, lparam)
   
      Wrapper of BroadcastMessage K-Meleon API.
   
   msg : number
   wparam : number
   lparam : number
 */
int km_BroadcastMessage(lua_State *L)
{
	UINT msg;
	WPARAM wparam;
	LPARAM lparam;
	
	luaL_checkany(L, 3);
	msg = luaL_checkinteger(L, 1);
	wparam = luaL_checkinteger(L, 2);
	lparam = luaL_checkinteger(L, 3);
	try {
		kFuncs->BroadcastMessage(msg, wparam, lparam);
	} catch (...) {
		RAISE_ERROR;
	}
	return 0;
}


/**
   ParseAccel(accel)
   
      Wrapper of ParseAccel K-Meleon API.
   
   accel : string : Accelerator string.
 */
int km_ParseAccel(lua_State *L)
{
	char *str;
	
	luaL_checkany(L, 1);
	str = lut_checkstring(L, 1);
	try {
		kFuncs->ParseAccel(str);
	} catch (...) {
		RAISE_ERROR;
	}
	return 0;
}


/**
   id(ID)
   
      Send a message id to the window.
   
   ID : string | number : Message ID.
 */
int km_id(lua_State *L)
{
	HWND hwnd;
	int id = -1;
	
	luaL_checkany(L, 1);
	hwnd = get_hwnd(L);
	id = get_commandid(L, 1);
	if (id > 0) {
		SendMessage(hwnd, WM_COMMAND, id, (LPARAM)NULL);
	}
	return 0;
}


/**
   GetFID(name [, arg] [, func])
   RegisterCommand(name [, arg] [, func])
   
      Gets funcfion command ID / Registers function as command.
      RegisterCommand is alias of GetFID for compatibility.
   
   name : string    : identifier
   arg  : string    : argument
   func : functiion : function for registration
   
   return : number : command ID or -1 (failed)
   
   Examples
   
      GetFID(name)
         Returns command ID corresponded to given notation
         that is same as GetID without plugin name.
         If a function given name is already registered as command,
         this API returns its ID.
   
      GetFID(name, arg)
         Returns command ID corresponded to given name and argument.
         If a function given name is already registered as command,
         this API returns its ID.
   
      GetFID(name, function)
         Regists function as command with given notation
         and returns own command ID.
         If a function given name is already registered,
         this API replaces the function by given function.
   
      GetFID(name, arg, function)
         Regists function as command with given name and argument,
         and returns own command ID.
         If a function given name is already registered,
         this API replaces the function by given function.
*/
int km_RegisterCommand(lua_State *L)
{
	static const char nil[] = "";
	char *param, *macro, *arg, *str;
	int id = -1, ref = LUA_NOREF;
	
	lua_settop(L, 3);
	param = lut_checkstring(L, 1);
	lua_insert(L, 1);
	switch (lua_type(L, 3)) {
	case LUA_TNIL:
		if (! parse_plugin_param(param, &macro, &arg, &str)) {
			lua_pushinteger(L, id);
			return 1;
		}
		break;
	case LUA_TSTRING:
		macro = param;
		arg = (char *) lua_tostring(L, 3);
		switch (lua_type(L, 4)) {
		case LUA_TNIL:
			break;
		case LUA_TFUNCTION:
			ref = lua_ref(L, 1);
			break;
		default:
			return luaL_argerror(L, 3, "invalid type");
		}
		break;
	case LUA_TFUNCTION:
		if (! parse_plugin_param(param, &macro, &arg, &str)) {
			lua_pushinteger(L, id);
			return 1;
		}
		lua_pop(L, 1);
		ref = lua_ref(L, 1);
		break;
	default:
		return luaL_argerror(L, 2, "invalid type");
	}
	if (! arg) arg = (char *) nil;
	if (ref == LUA_NOREF) {
		id = g_lpMacroset->search(macro, arg);
		if (id < 0) {
			GetRefParam par;
			par.macro = macro;
			if (lua_cpcall(L, get_function_ref, &par)) {
				return lua_error(L);
			} else {
				if (par.ref != LUA_NOREF) {
					id = g_lpMacroset->add(macro, arg, par.ref);
				}
			}
		}
	} else {
		id = g_lpMacroset->replace(macro, arg, &ref);
		if (id >= 0) {
			lua_unref(L, ref);
		}
	}
	lua_pushinteger(L, id);
	return 1;
}


#if (KMEL_PLUGIN_VER) > 0x0201 /* 1.0 API */

/**
   (1.0 feature)
   
   ToggleSideBar(index)
   
      Wrapper of ToggleSideBar K-Meleon API.
   
   index : number : 
*/
int km_ToggleSideBar(lua_State *L)
{
	int index;
	HWND hwnd;
	
	index = luaL_checkinteger(L, 1);
	hwnd = get_hwnd(L);
	if (hwnd) {
		try {
			kFuncs->ToggleSideBar(hwnd, index);
		} catch (...) {
			RAISE_ERROR;
		}
	}
	return 0;
}

/**
   (1.0 feature)
   
   InjectJS(script, top)
   
      Wrapper of InjectJS K-Meleon API.
      Inject JavaScript to document.
   
   script : string : 
   top    : boolean : 
   
   return : boolean : result
*/
int km_InjectJS(lua_State *L)
{
	const char *scr;
	bool top;
	HWND hwnd;
	BOOL result;
	
	luaL_checkany(L, 2);
	scr = lut_checkstring(L, 1);
	luaL_checktype(L, 2, LUA_TBOOLEAN);
	top = lua_toboolean(L, 2);
	hwnd = get_hwnd(L);
	try {
		result = kFuncs->InjectJS(scr, top, hwnd);
	} catch (...) {
		RAISE_ERROR;
	}
	lua_pushboolean(L, result);
	return 1;
}

/**
   (1.0 feature)
   
   InjectCSS(script, top)
   
      Wrapper of InjectCSS K-Meleon API.
      Inject CSS to document.
   
   script : string : 
   top    : boolean : 
   
   return : boolean : result
*/
int km_InjectCSS(lua_State *L)
{
	const char *scr;
	bool top;
	HWND hwnd;
	BOOL result;
	
	luaL_checkany(L, 2);
	scr = lut_checkstring(L, 1);
	luaL_checktype(L, 2, LUA_TBOOLEAN);
	top = lua_toboolean(L, 2);
	hwnd = get_hwnd(L);
	try {
		result = kFuncs->InjectCSS(scr, top, hwnd);
	} catch (...) {
		RAISE_ERROR;
	}
	lua_pushinteger(L, result);
	return 1;
}

/**
   (1.0 feature)
   
   GetInfoAtClick()
   
      Wrapper of GetInfoAtClick K-Meleon API.
   
   return : table :
   
   member of return value
   image : string : url of last clicked image
   link  : string : url of last clicked link
   frame : string : url of last clicked frame
   page  : string : url of last clicked page
*/
int km_GetInfoAtClick(lua_State *L)
{
	HWND hwnd;
	kmeleonPointInfo *pinfo;
	
	hwnd = get_hwnd(L);
	try {
		pinfo = kFuncs->GetInfoAtClick(hwnd);
	} catch (...) {
		RAISE_ERROR;
	}
	lua_newtable(L);
	if (pinfo->image) {
		lua_pushstring(L, pinfo->image);
		lua_setfield(L, -2, "image");
	}
	if (pinfo->link) {
		lua_pushstring(L, pinfo->link);
		lua_setfield(L, -2, "link");
	}
	if (pinfo->frame) {
		lua_pushstring(L, pinfo->frame);
		lua_setfield(L, -2, "frame");
	}
	if (pinfo->page) {
		lua_pushstring(L, pinfo->page);
		lua_setfield(L, -2, "page");
	}
	return 1;
}

/**
   (1.0 feature)
   
   GetKmeleonVersion()
   
      Wrapper of GetKmeleonVersion K-Meleon API.
   
   return : number : K-Meleon version
*/
int km_GetKmeleonVersion(lua_State *L)
{
	int ver;
	
	try {
		ver = kFuncs->GetKmeleonVersion();
	} catch (...) {
		RAISE_ERROR;
	}
	lua_pushinteger(L, ver);
	return 1;
}

/**
   (1.0 feature)
   
   Translate(text)
   _(text)
   
      translate text.
   
   text : string : 
   
   return : string : translated text
*/
int km_Translate(lua_State *L)
{
	const char *a;
	char *b;
	
	luaL_checkany(L, 1);
	b = lut_checkstring(L, 1);
	try {
		a = kFuncs->Translate(b);
	} catch (...) {
		RAISE_ERROR;
	}
	lua_pushstring(L, a);
	return 1;
}

/**
   (1.0 feature)
   
   SetGlobalVar(type, preference, value)
   
      Wrapper of SetGlobalVar K-Meleon API.
   
   type : TYPE_BOOL | TYPE_INT | TYPE_STRING : specify preference type
   preference : string : name of preference.
   value : boolean | number | string
   
   Below preferences can accept.
      TYPE_STRING : URLBAR TITLE CHARSET
      TYPE_INT    : TextZoom
   
   return : boolean : success or failure
 */
int km_SetGlobalVar(lua_State *L)
{
	PREFTYPE type;
	const char *pref;
	char *cValue;
	int nValue = 0;
	int ret;
	
	luaL_checkany(L, 3);
	type = (PREFTYPE) luaL_checkinteger(L, 1);
	pref = luaL_checkstring(L, 2);
	switch (type) {
	case PREF_BOOL:
		luaL_checktype(L, 3, LUA_TBOOLEAN);
		nValue = lua_toboolean(L, 3);
		try {
			ret = kFuncs->SetGlobalVar(type, pref, &nValue);
		} catch (...) {
			RAISE_ERROR;
		}
		break;
	case PREF_INT:
		nValue = luaL_checkinteger(L, 3);
		try {
			ret = kFuncs->SetGlobalVar(type, pref, &nValue);
		} catch (...) {
			RAISE_ERROR;
		}
		break;
	case PREF_STRING:
		cValue = lut_checkstring(L, 3);
		try {
			ret = kFuncs->SetGlobalVar(type, pref, cValue);
		} catch (...) {
			RAISE_ERROR;
		}
		break;
	default:
		return luaL_argerror(L, 1, "invalid pref type");
	}
	lua_pushboolean(L, ret);
	return 1;
}

#endif /* KMEL_PLUGIN_VER > 0x0201 */

#if (KMEL_PLUGIN_VER) > 0x0202 /* >= 1.1 */

/**
   (1.1 feature)
   
   GetFolder(type)
   
      Wrapper of GetFolder K-Meleon API.
   
   type : RootFolder | DefSettingsFolder | UserSettingsFolder |
          ProfileFolder | PluginsFolder | UserPluginsFolder |
          SkinsFolder | UserSkinsFolder | ResFolder | CurrentSkinFolder
   
   return : string
 */
int km_GetFolder(lua_State *L)
{
	enum FolderType type;
	char *path;
	long length;
	
	luaL_checkany(L, 1);
	type = (enum FolderType) luaL_checkinteger(L, 1);
	try {
		length = kFuncs->GetFolder(type, NULL, 0) + 1;
	} catch (...) {
		RAISE_ERROR;
	}
	path = (char *) lua_newuserdata(L, length);
	try {
		kFuncs->GetFolder(type, path, length);
	} catch (...) {
		RAISE_ERROR;
	}
	lua_pushstring(L, path);
	return 1;
}

/**
   (1.1 feature)
   
   SetAccel(key, command)
   
      Wrapper of SetAccel K-Meleon API.
   
   key     : string :
   command : string :
 */
int km_SetAccel(lua_State *L)
{
	char *key;
	char *command;
	
	luaL_checkany(L, 2);
	key = lut_checkstring(L, 1);
	command = lut_checkstring(L, 2);
	try {
		kFuncs->SetAccel(key, command);
	} catch (...) {
		RAISE_ERROR;
	}
	return 0;
}

/**
   (1.1 feature)
   
   SetMenu(name, type [, label] [, command] [, where])
   
      Wrapper of SetAccel K-Meleon API.
   
   name    : string :
   type    : MENU_COMMAND | MENU_POPUP | MENU_INLINE | MENU_PLUGIN |
             MENU_SEPARATOR : type of menu item
   label   : string : requires when type is which MENU_COMMAND,
                      MENU_POPUP, MENU_INLINE, MENU_PLUGIN.
   command : string : requires when type is MENU_COMMAND.
   where   : string | number : insert position
 */
int km_SetMenu(lua_State *L)
{
	const char *name;
	kmeleonMenuItem item = {0};
	int before = 3;
	
	name = luaL_checkstring(L, 1);
	item.command = 1;
	item.type = (short) luaL_checkinteger(L, 2);
	switch (item.type) {
	case MENU_SEPARATOR:
		item.label = "--"; /* needs something string */
		break;
	case MENU_INLINE:
	case MENU_PLUGIN:
	case MENU_POPUP:
		item.label = luaL_checkstring(L, 3);
		before = 4;
		break;
	case MENU_COMMAND:
		item.label = luaL_checkstring(L, 3);
		item.command = get_commandid(L, 4);
		if (item.command < 0) return 0;
		before = 5;
		break;
	default:
		return luaL_argerror(L, 2, "invalid menu type");
	}
	switch (lua_type(L, before)) {
	case LUA_TNUMBER:
		item.before = lua_tointeger(L, before);
		break;
	case LUA_TSTRING:
		item.before = (long) lua_tostring(L, before);
		break;
	default:
		item.before = -1;
		break;
	}
	try {
		kFuncs->SetMenu(name, &item);
	} catch (...) {
		RAISE_ERROR;
	}
	return 0;
}

/**
   (1.1 feature)
   
   RebuildMenu(name)
   
      Wrapper of RebuildMenu K-Meleon API.
   
   name : string :
 */
int km_RebuildMenu(lua_State *L)
{
	const char *menu = luaL_checkstring(L, 1);
	try {
		kFuncs->RebuildMenu(menu);
	} catch (...) {
		RAISE_ERROR;
	}
	return 0;
}

/**
   (1.1 feature)
   
   GetWindowVar(type)
   
      Wrapper of GetWindowVar K-Meleon API.
   
   type : Window_UrlBar | Window_Charset | Window_Title |
          Window_TextZoom | Window_URL |
          Window_SelectedText | Window_LinkURL |
          Window_ImageURL | Window_FrameURL
   
   result : string | number : 
 */
int km_GetWindowVar(lua_State *L)
{
	HWND hwnd;
	enum WindowVarType type;
	
	hwnd = get_hwnd(L);
	type = (enum WindowVarType) luaL_checkinteger(L, 1);
	switch (type) {
	case Window_TextZoom:
		{
			int zoom;
			try {
				kPlugin.kFuncs->GetWindowVar(hwnd, type, &zoom);
			} catch (...) {
				RAISE_ERROR;
			}
			lua_pushinteger(L, zoom);
		}
		break;
	case Window_UrlBar:
	case Window_Charset:
	case Window_Title:
	case Window_URL:
	case Window_SelectedText:
	case Window_LinkURL:
	case Window_ImageURL:
	case Window_FrameURL:
		{
			int length;
			try {
				length = kPlugin.kFuncs->GetWindowVar(hwnd, type, NULL);
			} catch (...) {
				RAISE_ERROR;
			}
			if (length < 1) {
				lua_pushstring(L, "");
			} else {
				char *result;
				result = (char *) lua_newuserdata(L, length);
				try {
					kPlugin.kFuncs->GetWindowVar(hwnd, type, (void *) result);
				} catch (...) {
					RAISE_ERROR;
				}
				lua_pushstring(L, result);
			}
		}
		break;
	default:
		return luaL_argerror(L, 1, "invalid type");
	}
	return 1;
}

/**
   (1.1 feature)
   
   SetWindowVar(type, value)
   
      Wrapper of SetWindowVar K-Meleon API.
   
   type  : Window_UrlBar | Window_Charset | Window_Title |
           Window_TextZoom | Window_URL
   value : string | number
   
   result : boolean
 */
int km_SetWindowVar(lua_State *L)
{
	HWND hwnd;
	enum WindowVarType type;
	BOOL result = FALSE;
	
	hwnd = get_hwnd(L);
	type = (enum WindowVarType) luaL_checkinteger(L, 1);
	switch (type) {
	case Window_TextZoom:
		{
			int zoom;
			zoom = luaL_checkinteger(L, 2);
			try {
				result = kPlugin.kFuncs->SetWindowVar(hwnd, type, &zoom);
			} catch (...) {
				RAISE_ERROR;
			}
		}
		break;
	case Window_UrlBar:
	case Window_Charset:
	case Window_Title:
	case Window_URL:
		{
			const char *value;
			value = luaL_checkstring(L, 2);
			try {
				result = kPlugin.kFuncs->SetWindowVar(hwnd, type, (void *) value);
			} catch (...) {
				RAISE_ERROR;
			}
		}
		break;
	default:
		return luaL_argerror(L, 1, "invalid type");
	}
	lua_pushboolean(L, result);
	return 1;
}

/**
   (1.1 feature)
   
   GetMozillaSessionHistory()
   
      Wrapper of GetMozillaSessionHistory K-Meleon API.
   
   result :
      titles | table  |
      urls   | table  |
      index  | number | starts at 1.
 */
int km_GetMozillaSessionHistory(lua_State *L)
{
	HWND hwnd;
	char **titles, **urls;
	int count, index, i;
	
	lua_settop(L, 0);
	hwnd = get_hwnd(L);
	try {
		if (! kFuncs->GetMozillaSessionHistory(hwnd, &titles, &urls,
		                                       &count, &index)) {
			return 0;
		}
	} catch (...) {
		RAISE_ERROR;
	}
	lua_createtable(L, 0, count); /* titles 1 */
	lua_createtable(L, 0, count); /* urls 2 */
	for (i = 0; i < count; ++ i) {
		lua_pushstring(L, titles[i]);
		lua_rawseti(L, 1, i + 1);
		lua_pushstring(L, urls[i]);
		lua_rawseti(L, 2, i + 1);
	}
	lua_pushinteger(L, index + 1);
	return 3;
}

/**
   (1.1 feature)
   
   SetMozillaSessionHistory(titles, urls, index)
   
      Wrapper of SetMozillaSessionHistory K-Meleon API.
   
   titles | table  |
   urls   | table  |
   index  | number | starts at 1.
   
   result : boolean
 */
int km_SetMozillaSessionHistory(lua_State *L)
{
	HWND hwnd;
	char **titles, **urls;
	const char *s;
	int count, index, result = 0, err = FALSE, i;
	
	hwnd = get_hwnd(L);
	luaL_checktype(L, 1, LUA_TTABLE);
	luaL_checktype(L, 2, LUA_TTABLE);
	index = luaL_checkinteger(L, 3);
	
	count = (int) lua_objlen(L, 1);
	if (count < 1 || index < 0 || count != (int) lua_objlen(L, 2)) {
		lua_pushboolean(L, result);
		return 1;
	}
	titles = (char **) malloc(sizeof(char **) * count);
	urls = (char **) malloc(sizeof(char **) * count);
	if (titles && urls) {
		for (i = 0; i < count; ++ i) {
			titles[i] = urls[i] = NULL;
			
			lua_rawgeti(L, 1, i + 1);
			if (! (s = lua_tostring(L, -1))) break;
			titles[i] = _strdup(s);
			lua_pop(L, 1);
			
			lua_rawgeti(L, 2, i + 1);
			if (! (s = lua_tostring(L, -1))) break;
			urls[i] = _strdup(s);
			lua_pop(L, 1);
			
			if (! titles[i] || ! urls[i]) break;
		}
		if (i == count) {
			try {
				result = kFuncs->SetMozillaSessionHistory(hwnd,
				                                          (const char **) titles,
				                                          (const char **) urls,
				                                          count, index - 1);
			} catch (...) {
				err = TRUE;
			}
			-- i;
		}
		for (;i >= 0; -- i) {
			free(titles[i]);
			free(urls[i]);
		}
	}
	free(titles);
	free(urls);
	if (err) RAISE_ERROR;
	lua_pushboolean(L, result);
	return 1;
}

#endif /* KMEL_PLUGIN_VER > 0x0202 */



/* functions */
const luaL_Reg funcs[] = {
	{"plugin", km_plugin},
	{"pluginmsg", km_pluginmsg},
	{"pluginmsgex", km_pluginmsgex},
	{"GetCommandIDs", km_GetCommandIDs},
	{"NavigateTo", km_NavigateTo},
	{"GetDocInfo", km_GetDocInfo},
	{"getpref", km_getpref},
	{"setpref", km_setpref},
	{"delpref", km_delpref},
	{"SetStatusBarText", km_SetStatusBarText},
	{"GotoHistoryIndex", km_GotoHistoryIndex},
	{"GetID", km_GetID},
	{"GetGlobalVar", km_GetGlobalVar},
	{"EncodeUTF8", km_EncodeUTF8},
	{"DecodeUTF8", km_DecodeUTF8},
	{"GetMenu", km_GetMenu},
	{"SetForceCharset", km_SetForceCharset},
	{"SetCheck", km_SetCheck},
	{"ClearCache", km_ClearCache},
	{"BroadcastMessage", km_BroadcastMessage},
	{"ParseAccel", km_ParseAccel},
	{"id", km_id},
	{"RegisterCommand", km_RegisterCommand},
	{"GetFID", km_RegisterCommand},
#if (KMEL_PLUGIN_VER) > 0x0201 /* 1.0 API */
	{"ToggleSideBar", km_ToggleSideBar},
	{"InjectJS", km_InjectJS},
	{"InjectCSS", km_InjectCSS},
	{"GetInfoAtClick", km_GetInfoAtClick},
	{"GetKmeleonVersion", km_GetKmeleonVersion},
	{"Translate", km_Translate},
	{"_", km_Translate},
	{"SetGlobalVar", km_SetGlobalVar},
#endif /* KMEL_PLUGIN_VER > 0x0201 */
#if (KMEL_PLUGIN_VER) > 0x0202 /* 1.1 API */
	{"GetFolder", km_GetFolder},
	{"SetAccel", km_SetAccel},
	{"SetMenu", km_SetMenu},
	{"RebuildMenu", km_RebuildMenu},
	{"GetWindowVar", km_GetWindowVar},
	{"SetWindowVar", km_SetWindowVar},
	{"GetMozillaSessionHistory", km_GetMozillaSessionHistory},
	{"SetMozillaSessionHistory", km_SetMozillaSessionHistory},
#endif /* KMEL_PLUGIN_VER > 0x0202 */
	/* end */
	{NULL, 0},
};

int luaopen_km(lua_State *L)
{
	const char *libname;
	
	libname = luaL_optstring(L, 1, "_G");
	/*
	  set constants and functions
	*/
	/* k-meleon api */
	luaL_register(L, libname, funcs);
	lut_registint(L, consts, -1);
	/* win32 */
	luaL_register(L, NULL, w32_funcs);
	lut_registint(L, w32_consts, -1);
	return 1;
}
