/*
 *  Copyright (C) 2005-2006 mark307
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <shellapi.h>
#include <stdlib.h>

#define KMELEON_PLUGIN_EXPORTS
#include <kmeleon_plugin.h>
#include <KMeleonConst.h>

extern "C" {
#include <lua.h>
#include <lualib.h>
#include <lauxlib.h>
}

/* TAB Messages */
#ifndef WM_CLOSETAB
#define WM_CLOSETAB WM_APP + 150
#define WM_CREATETAB WM_APP + 151
#define WM_SWITCHTAB WM_APP + 152
#endif

#include "luamacros.h"

#include "macroset.h"
#include "kmapi.h"
#include "kmenv.h"

#define set_hook(L) lua_sethook(L, L_CountHook, LUA_MASKCOUNT, HOOK_COUNT)
#define remove_hook(L) lua_sethook(L, NULL, 0, 0)

typedef struct {
	union {
		const char *name;
		int id;
	};
	HWND hwnd;
	lua_Integer arg;
} CallbackParam;

typedef struct {
	union {
		const char *name;
		int id;
	};
	HWND hwnd;
	const char *arg;
	char *result;
} CallbackParam2;


long DoMessage(const char *, const char *, const char *, long, long);

LRESULT CALLBACK WndProc(HWND, UINT, WPARAM, LPARAM);
static int Load();
static void Create(HWND);
static void Config(HWND);
static void CloseFrame(HWND, LONG);
static void DoMenu(HMENU, char *);
static void DoRebar(HWND);
static int DoAccel(char *);
static void Exit();
static char *FindNick(const char *);
static void CallHook(const char *, HWND, lua_Integer);
static char *CallHook2(const char *, HWND, const char *);

kmeleonPlugin kPlugin = {
	KMEL_PLUGIN_VER,
	PLUGIN_NAME,
	DoMessage
};

configFileType g_configFiles[1] = {
	{"Lua", CONFIG_NAME,  "http://kmeleon.sourceforge.net/"},
};

kmeleonFunctions *kFuncs;
WNDPROC KMeleonWndProc;
int wm_deferopenmsg = -1;
int g_bStartup = 1;

MacroSet* g_lpMacroset;
CRITICAL_SECTION g_hCriticalSection;
lua_State *g_luaState;




extern "C" {
	BOOL WINAPI DllMain(HINSTANCE hInstance, DWORD dwReason, LPVOID lpReserved)
	{
		switch (dwReason) {
		case DLL_PROCESS_ATTACH:
			InitializeCriticalSection(&g_hCriticalSection);
			g_lpMacroset = new MacroSet();
			return g_lpMacroset != NULL;
		case DLL_PROCESS_DETACH:
			if (g_luaState) {
				lua_close(g_luaState);
			}
			if (g_lpMacroset) {
				delete g_lpMacroset;
			}
			DeleteCriticalSection(&g_hCriticalSection);
		}
		return TRUE;
	}
	KMELEON_PLUGIN kmeleonPlugin *GetKmeleonPlugin() {
		return &kPlugin;
	}
	KMELEON_PLUGIN int DrawBitmap(DRAWITEMSTRUCT *dis) {
		return 14; // 14 = icon width
	}
}


static int call_lua_callback(lua_State *L)
{
	CallbackParam *p = (CallbackParam *) lua_touserdata(L, 1);
	
	set_hwnd(L, p->hwnd);
	lua_pushstring(L, p->name);
	lua_gettable(L, LUA_GLOBALSINDEX);
	if (lua_isfunction(L, -1)) {
		remove_hook(L);
		set_hook(L);
		lua_pushinteger(L, p->arg);
		lua_call(L, 1, 1);
	}
	return 0;
}


static void CallHook(const char *funcname, HWND hwnd, lua_Integer arg)
{
	if (g_luaState) {
		CallbackParam p;
		
		EnterCriticalSection(&g_hCriticalSection);
		p.name = funcname;
		p.hwnd = hwnd;
		p.arg = arg;
		if (lua_cpcall(g_luaState, call_lua_callback, &p)) {
			L_DoError(g_luaState);
			lua_pop(g_luaState, 1);
		}
		LeaveCriticalSection(&g_hCriticalSection);
	}
}


static int call_lua_callback2(lua_State *L)
{
	CallbackParam2 *p = (CallbackParam2 *) lua_touserdata(L, 1);
	
	set_hwnd(L, p->hwnd);
	lua_pushstring(L, p->name);
	lua_gettable(L, LUA_GLOBALSINDEX);
	if (lua_isfunction(L, -1)) {
		remove_hook(L);
		set_hook(L);
		lua_pushstring(L, p->arg);
		lua_call(L, 1, 1);
		if (lua_isstring(L, -1)) {
			p->result = _strdup(lua_tostring(L, -1));
		}
	}
	return 0;
}


static char *CallHook2(const char *funcname, HWND hwnd, const char *arg)
{
	CallbackParam2 p;
	
	p.result = NULL;
	if (g_luaState) {
		
		EnterCriticalSection(&g_hCriticalSection);
		p.name = funcname;
		p.hwnd = hwnd;
		p.arg = arg;
		if (lua_cpcall(g_luaState, call_lua_callback2, &p)) {
			L_DoError(g_luaState);
			lua_pop(g_luaState, 1);
		}
		LeaveCriticalSection(&g_hCriticalSection);
	}
	return p.result;
}


long DoMessage(const char *to, const char *from, const char *subject,
			   long data1, long data2)
{
	if (to[0] == '*' || stricmp(to, kPlugin.dllname) == 0) {
		if (stricmp(subject, "Load") == 0) {
			Load();
		}
		else if (stricmp(subject, "CloseFrame") == 0) {
			CloseFrame((HWND) data1, data2);
		}
		else if (stricmp(subject, "Config") == 0) {
			Config((HWND) data1);
		}
		else if (stricmp(subject, "Create") == 0) {
			Create((HWND) data1);
		}
		else if (stricmp(subject, "DoMenu") == 0) {
			DoMenu((HMENU) data1, (char *) data2);
		}
		else if (stricmp(subject, "DoRebar") == 0) {
			DoRebar((HWND) data1);
		}
		else if (stricmp(subject, "DoAccel") == 0) {
			*(int *)data2 = DoAccel((char *) data1);
		}
		else if (stricmp(subject, "Exit") == 0) {
			Exit();
		}
		else if (stricmp(subject, "FindNick") == 0) {
			*(char **)data2 = FindNick((char *) data1);
		}
		else if (stricmp(subject, "GetConfigFiles") == 0) {
			*(int *)data2 = 1;
			data1 = (long) g_configFiles;
		}
		else if (stricmp(subject, "Close") == 0) {
			CallHook("OnCloseWindow", (HWND) data1, 0);
		}
		else if (stricmp(subject, "Destroy") == 0) {
			CallHook("OnDestroy", (HWND) data1, 0);
		}
		else if (stricmp(subject, "Init") == 0) {
			CallHook("OnInit", NULL, 0);
		}
		else if (stricmp(subject, "Quit") == 0) {
			CallHook("OnQuit", NULL, 0);
		}
		else if (stricmp(subject, "Setup") == 0) {
			CallHook("OnSetup", NULL, 0);
		}
		else return 0;
		return 1;
	}
	return 0;
}


static void get_browser_path(lua_State *L, char *szPath, int size)
{
	char *p;
	
	szPath[0] = '\0';
	if (GetModuleFileName(kPlugin.hParentInstance, szPath, size)) {
		p = szPath + lstrlen(szPath);
		/* It depends on browser's filename to contain only ASCII */
		while (*p != '\\' && p > szPath) {
			*p = '\0';
			p --;
		}
	}
}


static int setup_lua(lua_State *L)
{
	static const luaL_Reg lualibs[] = {
		{"", luaopen_base},
		{LUA_TABLIBNAME, luaopen_table},
		{LUA_STRLIBNAME, luaopen_string},
		{LUA_LOADLIBNAME, luaopen_package},
		{LUA_IOLIBNAME, luaopen_io},
		{LUA_OSLIBNAME, luaopen_os},
		{LUA_MATHLIBNAME, luaopen_math},
		{LUA_DBLIBNAME, luaopen_debug},
		{NULL, NULL}
	};
	const luaL_Reg *lib = lualibs;
	char **szPath;
	
	szPath = (char **) lua_touserdata(L, 1);
	/*
	 * open lua library
	 */
	for (; lib->func; lib ++) {
		lua_pushcfunction(L, lib->func);
		lua_pushstring(L, lib->name);
		lua_call(L, 1, 0);
	}
	/*
	 * tweak lua environment
	 */
	/* override os.exit */
	lua_getfield(L, LUA_GLOBALSINDEX, LUA_OSLIBNAME);
	if (lua_istable(L, -1)) {
		lua_pushcfunction(L, L_DoExit);
		lua_setfield(L, -2, "exit");
	}
	lua_pop(L, 1);
	/* override print */
	lua_pushcfunction(L, L_DoPrint);
	lua_setfield(L, LUA_GLOBALSINDEX, "print");
	/* set package loadpath */
	lua_getfield(L, LUA_GLOBALSINDEX, "package");
	if (lua_istable(L, -1)) {
		lua_pushfstring(L, "%s?.lua;%s?.lc", szPath[1], szPath[1]);
		lua_setfield(L, -2, "path");
		lua_pushfstring(L, "%s?.dll", szPath[1]);
		lua_setfield(L, -2, "cpath");
	}
	lua_pop(L, 1);
	/*
	 * initialize K-Meleon API
	 */
	lua_pushcfunction(L, luaopen_km);
	lua_pushstring(L, LUA_KMELEONLIBNAME);
	lua_call(L, 1, 1);
	/*
	 * set convinience variables
	 */
	/* PREFPATH */
	lua_pushstring(L, szPath[0]);
	lua_setfield(L, -2, "PREFPATH");
	/* LIBPATH */
	lua_pushstring(L, szPath[1]);
	lua_setfield(L, -2, "LIBPATH");
	/* BROWSERPATH */
	get_browser_path(L, szPath[2], MAX_PATH);
	lua_pushstring(L, szPath[2]);
	lua_setfield(L, -2, "BROWSERPATH");
	/*
	 * set global metatable
	 * 
	 *  global table has km table.
	 *  global table inherits km by metamethod __index.
	 *  see the implement of module.seeall (in loadlib.c)
	 */
	if (! lua_getmetatable(L, LUA_GLOBALSINDEX)) {
		lua_createtable(L, 0, 1);
		lua_setmetatable(L, LUA_GLOBALSINDEX);
		if (! lua_getmetatable(L, LUA_GLOBALSINDEX)) {
			return luaL_error(L, "Internal error.");
		}
	}
	/* swap metatable and km table */
	lua_insert(L, 2);
	/* inherit global from km */
	lua_setfield(L, -2, "__index");
	/* pop global metatable */
	lua_pop(L, 1);
	/* instance handle of browser */
	set_parentinstance(L, kPlugin.hParentInstance);
	/* instance handle of this plugin */
	set_dllinstance(L, kPlugin.hDllInstance);
	/*
	 * load and eval initialize script
	 */
	remove_hook(L);
	set_hook(L);
	if (luaL_dofile(L, g_configFiles[0].file)) {
		return lua_error(L);
	}
	return 0;
}


static int Load()
{
	char szPrefPath[MAX_PATH], szLibPath[MAX_PATH];
	char *param[3];
	
	kFuncs = kPlugin.kFuncs;
	
	/*
	 * check string length
	 */
#if (KMEL_PLUGIN_VER) > 0x0201
	{
		long length;
		length = kFuncs->GetPreference(PREF_STRING, LIBPATH,
									   NULL, (void *) "");
		if (length > MAX_PATH) return 1;
#if (KMEL_PLUGIN_VER) == 0x0202
		length = kFuncs->GetPreference(PREF_STRING, PREFPATH,
									   NULL, (void *) "");
		if (length > MAX_PATH) return 1;
#else
		length = kFuncs->GetFolder(UserSettingsFolder, NULL, 0);
		length ++;
		if (length > MAX_PATH) return 1;
#endif
	}
#endif
	/*
	 * set config path
	 */
	szPrefPath[0] = szLibPath[0] = '\0';
#if (KMEL_PLUGIN_VER) < 0x0203
	// potential vulnerability under 1.0
	kFuncs->GetPreference(PREF_STRING, PREFPATH, szPrefPath, (void *) "");
#else
	kFuncs->GetFolder(UserSettingsFolder, szPrefPath, MAX_PATH);
	strcat(szPrefPath, "\\");
#endif
	kFuncs->GetPreference(PREF_STRING, LIBPATH, szLibPath, (void *) "");
	/* */
	param[0] = szPrefPath;
	param[1] = (! szLibPath[0]) ? szPrefPath : szLibPath;
	param[2] = szLibPath;
	/*
	 * 
	 */
	EnterCriticalSection(&g_hCriticalSection);
	/*
	 * make initialize script path
	 */
	if (strlen(szPrefPath) + strlen(CONFIG_NAME) + 1 < sizeof(g_configFiles[0].file)) {
		strcpy(g_configFiles[0].file, szPrefPath);
		strcat(g_configFiles[0].file, CONFIG_NAME);
	}
	/*
	 * initialize lua library
	 */
	if (g_luaState) {
		LeaveCriticalSection(&g_hCriticalSection);
		return 0;
	}
	g_luaState = luaL_newstate();
	if (g_luaState) {
		/*
		 * setup script environment
		 */
		if (lua_cpcall(g_luaState, setup_lua, param)) {
			L_DoError(g_luaState);
			lua_close(g_luaState);
			g_luaState = NULL;
		}
	}
	/*
	 * end of setup
	 */
	LeaveCriticalSection(&g_hCriticalSection);
	/*
	 * reserve private message
	 */
	wm_deferopenmsg = kPlugin.kFuncs->GetCommandIDs(1);
	return 0;
}


static void Create(HWND hWndParent)
{
	KMeleonWndProc = (WNDPROC) GetWindowLong(hWndParent, GWL_WNDPROC);
	SetWindowLong(hWndParent, GWL_WNDPROC, (LONG)WndProc);
	
	PostMessage(hWndParent, WM_COMMAND, wm_deferopenmsg, 0);
}


static void Config(HWND hWndParent)
{
	static const char *verb[] = {"edit", NULL};
	const char **p = verb;
	SHELLEXECUTEINFO sei = {0};
	
	/* try known verb */
	do {
		if ((int) ShellExecute(hWndParent, *p, g_configFiles[0].file,
		                       NULL, NULL, SW_SHOW) > 32) {
			return;
		}
	} while (*p != NULL && p ++);
	/* try ShellExecuteEx */
	sei.cbSize = sizeof(sei);
	sei.hwnd = hWndParent;
	sei.lpFile = g_configFiles[0].file;
	sei.nShow = SW_SHOW;
	if (ShellExecuteEx(&sei)) return;
	/* finally using notepad */
	ShellExecute(hWndParent, NULL, "notepad.exe", g_configFiles[0].file,
	             NULL, SW_SHOW);
}




static void CloseFrame(HWND hWnd, LONG windows)
{
	CallHook("OnCloseGroup", hWnd, (lua_Integer) windows);
}


static char *FindNick(const char *nick)
{
	return CallHook2("OnFindNick", NULL, nick);
}


static void Exit()
{
	CallHook("OnExit", NULL, 0);
	if (g_luaState) {
		lua_close(g_luaState);
		g_luaState = NULL;
	}
	if (g_lpMacroset) {
		delete g_lpMacroset;
		g_lpMacroset = NULL;
	}
}


static void DoMenu(HMENU hMenu, char *param)
{
	int id;
	char *macro, *arg, *str;
	static const char nil[] = "";
	
	if (! param || ! g_luaState) {
		return;
	}
	macro = arg = str = NULL;
	if (! parse_plugin_param(param, &macro, &arg, &str)) {
		return;
	}
	if (! arg) arg = (char *) nil;
	EnterCriticalSection(&g_hCriticalSection);
	id = g_lpMacroset->search(macro, arg);
	if (id < 0) {
		GetRefParam par;
		par.macro = macro;
		if (lua_cpcall(g_luaState, get_function_ref, &par)) {
			L_DoError(g_luaState);
			lua_pop(g_luaState, 1);
		} else {
			if (par.ref != LUA_NOREF) {
				id = g_lpMacroset->add(macro, arg, par.ref);
			}
		}
	}
	LeaveCriticalSection(&g_hCriticalSection);
	if (id > 0) {
		if (str) {
			AppendMenu(hMenu, MF_STRING, id, str);
		} else {
			AppendMenu(hMenu, MF_STRING, id, macro);
		}
	}
}


static void DoRebar(HWND hWndRebar)
{
}


static int DoAccel(char *param)
{
	int id;
	char *macro, *arg, *str;
	static const char nil[] = "";
	
	if (! param || ! g_luaState) {
		return -1;
	}
	macro = arg = str = NULL;
	if (! parse_plugin_param(param, &macro, &arg, &str)) {
		return -1;
	}
	if (! arg) arg = (char *) nil;
	EnterCriticalSection(&g_hCriticalSection);
	id = g_lpMacroset->search(macro, arg);
	if (id < 0) {
		GetRefParam par;
		par.macro = macro;
		if (lua_cpcall(g_luaState, get_function_ref, &par)) {
			L_DoError(g_luaState);
			lua_pop(g_luaState, 1);
		} else {
			if (par.ref != LUA_NOREF) {
				id = g_lpMacroset->add(macro, arg, par.ref);
			}
		}
	}
	LeaveCriticalSection(&g_hCriticalSection);
	return id;
}


int call_function(lua_State *L)
{
	CallbackParam *p = (CallbackParam *) lua_touserdata(L, 1);
	const MacroSet::Node *node = g_lpMacroset->search(p->id);
	if (node != NULL) {
		set_hwnd(L, p->hwnd);
		remove_hook(L);
		set_hook(L);
		lua_getref(L, node->ref);
		lua_pushstring(L, node->arg);
		lua_call(L, 1, 0);
	}
	return 0;
}


LRESULT CALLBACK WndProc(HWND hWnd, UINT message,
						 WPARAM wParam, LPARAM lParam)
{
	if (g_luaState) {
		switch (message) {
		case WM_COMMAND:
			if (LOWORD(wParam) == wm_deferopenmsg) {
				if (g_bStartup) {
					g_bStartup = 0;
					CallHook("OnStartup", hWnd, 0);
				}
				CallHook("OnOpenWindow", hWnd, 0);
			}
#ifdef ID_APP_EXIT
			else if (LOWORD(wParam) == ID_APP_EXIT) {
				CallHook("OnWMAppExit", hWnd, 0);
			}
#endif
#if 0
			else if (LOWORD(wParam) == WM_CLOSE) {
				CallHook("OnWMClose", hWnd, 0);
			}
#endif
			else {
				CallbackParam p;
				
				p.id = LOWORD(wParam);
				p.hwnd = hWnd;
				EnterCriticalSection(&g_hCriticalSection);
				if (lua_cpcall(g_luaState, call_function, &p)) {
					L_DoError(g_luaState);
					lua_pop(g_luaState, 1);
				}
				LeaveCriticalSection(&g_hCriticalSection);
			}
			break;
		case UWM_UPDATEBUSYSTATE:
			if (wParam == 0) {
				CallHook("OnLoad", hWnd, 0);
			}
			break;
#if (KMEL_PLUGIN_VER) > 0x0201
		case WM_CLOSETAB:
			CallHook("OnCloseTab", hWnd, (lua_Integer) lParam);
			break;
		case WM_CREATETAB:
			CallHook("OnCreateTab", hWnd, (lua_Integer) lParam);
			break;
		case WM_SWITCHTAB:
			CallHook("OnSwitchTab", hWnd, (lua_Integer) lParam);
			break;
		case WM_ACTIVATE:
			if ((LOWORD(wParam) == WA_ACTIVE ||
			     LOWORD(wParam) == WA_CLICKACTIVE) &&
			    (! lParam ||
			     GetClassLongPtr(hWnd, GCW_ATOM) ==
			     GetClassLongPtr((HWND) lParam, GCW_ATOM))) {
				LRESULT result = CallWindowProc(KMeleonWndProc, hWnd,
				                                message, wParam, lParam);
				CallHook("OnActivateWindow", hWnd, 0);
				return result;
			}
			break;
#endif
		}
	}
	return CallWindowProc(KMeleonWndProc, hWnd, message, wParam, lParam);
}
