--[[

  privacy.lua - alternative of privacy plugin
  
  This is example of using Lua addon.
  
  luamacros(privacy.ClearCookies)
  luamacros(privacy.ClearHistory)
  luamacros(privacy.ClearMRU)
  luamacros(privacy.ClearCache)
  luamacros(privacy.RemoveCache)
  luamacros(privacy.ClearSignon)

  Note: If you want to use RemoveCache, you have to install
        LuaFileSystem Addon (lfs.dll) into "package.cpath"
        (default is same as PREFPATH).
        http://luaforge.net/projects/luafilesystem
]]

pcall(require, "hook")
pcall(require, "lfs")

privacy = {
	ONSTARTUP = 1,
	ONSHUTDOWN = 2,
	ONBOTH = 3,
	
	prefClearCookies = 0,
	prefClearCache = 0,
	prefClearMRU = 0,
	prefClearSignon = 0,
	prefClearHistory = 0,
}

local pref = {
	settings_dir = "kmeleon.general.settingsDir",
	cache_parentdir = "browser.cache.disk.parent_directory",
	mru_maxurls = "kmeleon.MRU.maxURLs",
	mru_url = "kmeleon.MRU.URL",
	signon_file = "signon.SignonFileName",
	
	clearcookies = "kmeleon.plugins.luamacros.privacy.clearCookies",
	clearcache = "kmeleon.plugins.luamacros.privacy.clearCache",
	clearmru = "kmeleon.plugins.luamacros.privacy.clearMRU",
	clearsignon = "kmeleon.plugins.luamacros.privacy.clearSignOn",
	clearhistory = "kmeleon.plugins.luamacros.privacy.clearHistory",
}

local settingsDir, cacheParentDir, signonFileName

function privacy.LoadPreferences()
	settingsDir = getpref(TYPE_STRING, pref.settings_dir)
	cacheParentDir = getpref(TYPE_STRING, pref.cache_parentdir)
	signonFileName = getpref(TYPE_STRING, pref.signon_file)
	
	privacy.prefClearCookies = getpref(TYPE_INT, pref.clearcookies)
	privacy.prefClearCache = getpref(TYPE_INT, pref.clearcache)
	privacy.prefClearMRU = getpref(TYPE_INT, pref.clearmru)
	privacy.prefClearSignon = getpref(TYPE_INT, pref.clearsignon)
	privacy.prefClearHistory = getpref(TYPE_INT, pref.clearhistory)
end

function privacy.SavePreferences()
	setpref(TYPE_INT, pref.clearcookies, privacy.prefClearCookies)
	setpref(TYPE_INT, pref.clearcache, privacy.prefClearCache)
	setpref(TYPE_INT, pref.clearmru, privacy.prefClearMRU)
	setpref(TYPE_INT, pref.clearsignon, privacy.prefClearSignon)
	setpref(TYPE_INT, pref.clearhistory ,privacy.prefClearHistory)
end

function privacy.ClearCookies()
	os.remove(settingsDir .. "cookies.txt")
end

function privacy.ClearHistory()
	alert(settingsDir .. "history.txt")
	os.remove(settingsDir .. "history.txt")
end

function privacy.ClearSignon()
	os.remove(settingsDir .. signonFileName)
end

function privacy.ClearCache()
	ClearCache(STORE_ANYWHERE)
	ClearCache(STORE_IN_MEMORY)
	ClearCache(STORE_ON_DISK)
	ClearCache(STORE_ON_DISK_AS_FILE)
end

function privacy.RemoveCache()
	local k, v, i, f, a
	if lfs then
		for k, v in pairs({"Cache\\", "Cache.Trash\\"}) do
			for i in lfs.dir(cacheParentDir .. v) do
				f = cacheParentDir .. v .. i
				a = lfs.attributes(f)
				if a.mode == "file" then
					os.remove(f)
				end
			end
		end
	end
end

function privacy.ClearMRU()
	mrus = getpref(TYPE_INT, pref.mru_maxurls)
	for i = 0, mrus, 1 do
		setpref(TYPE_STRING, pref.mru_url .. i, "")
	end
	BroadcastMessage(UWM_REFRESHMRULIST, 0, 0)
end

function privacy.DoShutdownTasks()
	if privacy.prefClearCookies > 1 then privacy.ClearCookies() end
	if privacy.prefClearHistory > 1 then privacy.ClearHistory() end
	if privacy.prefClearMRU > 1 then privacy.ClearMRU() end
	if privacy.prefClearCache > 1 then
		privacy.ClearCache()
		privacy.RemoveCache()
	end
	if privacy.prefClearSignon > 1 then privacy.ClearSignon() end
end

function privacy.DoStartupTasks()
	if privacy.prefClearCookies % 2 == 1 then privacy.ClearCookies() end
	if privacy.prefClearHistory % 2 == 1 then privacy.ClearHistory() end
	if privacy.prefClearMRU % 2 == 1 then privacy.ClearMRU() end
	if privacy.prefClearCache % 2 == 1 then
		privacy.ClearCache()
		privacy.RemoveCache()
	end
	if privacy.prefClearSignon % 2 == 1 then privacy.ClearSignon() end
end

privacy.LoadPreferences()
privacy.SavePreferences()
privacy.DoStartupTasks()

if hook then
	hook.add(ExitHook, privacy.DoShutdownTasks)
end
