--[[    Simple search sample for 1.0

luamacros(search.Search, Search)
luamacros(search.SelectSearch, Search with)
--
luamacros(search.Search(Google), Google)
luamacros(search.Search(Yahoo), Yahoo)
luamacros(search.Search(Wikipedia), Wikipedia)
luamacros(search.Search(Dictionaries), Dictionaries)
luamacros(search.Search(KM), K-Meleon Forums)
luamacros(search.Search(URL), URL)
--
luamacros(search.SetSearch, Set Default Engine)
--
luamacros(search.SetOpen(0), Open results in current page);
luamacros(search.SetOpen(1), Open results in new page);
luamacros(search.SetOpen(2), Open results in background);

]]

if not KMEL_PLUGIN_VER or KMEL_PLUGIN_VER < 514 then
	return require "search"
end

require "hook"
require "layer"

search = {
	mes = {
		prompt_caption = "Search",
		prompt_prompt = "Search %s for:",
		select_prompt = "Select search engine",
		select_caption = "Change search engine",
		enter_url = "Enter URL:",
		go_url = "Go URL",
	},
	
	se = "kmeleon.plugins.luamacros.search.searchEngine",
	so = "kmeleon.plugins.luamacros.search.searchOpen",
	
	encode = function (query)
		return string.gsub(query, "([^a-zA-Z0-9_%.%-])", function (s)
			if s == " " then
				return "+"
			else
				return string.format("%%%02X", string.byte(s))
			end
		end)
	end,
	
	utf8encode = function (query)
		return search.encode(km.EncodeUTF8(query))
	end,
}

search.engines = {
	URL = {
		name = nil,
		url = "%s",
		filter = nil,
	},
	Google = {
		name = "Google",
		url = "http://www.google.com/search?q=%s",
		filter = search.utf8encode,
	},
	GoogleImages = {
		name = "Google Images",
		url = "http://images.google.com/images?btnG=Google+Search&safe=off&filter=0&q=%s",
		filter = search.utf8encode,
	},
	GoogleGroups = {
		name = "Google Groups",
		url = "http://groups.google.co.uk/groups?hl=en&btnG=Google+Search&q=%s",
		filter = search.utf8encode,
	},
	GoogleScholar = {
		name = "Google Scholar",
		url = "http://www.scholar.google.com/scholar?q=%s",
		filter = search.utf8encode,
	},
	MSNSearch = {
		name = "MSN Search",
		url = "http://search.msn.com/results.aspx?q=%s",
		filter = search.utf8encode,
	},
	Yahoo = {
		name = "Yahoo!",
		url = "http://search.yahoo.com/bin/search?p=%s",
		filter = search.encode,
	},
	Wikipedia = {
		name = "Wikipedia",
		url = "http://en.wikipedia.org/wiki/%s",
		filter = search.utf8encode,
	},
	OneLook = {
		name = "OneLook Dictionary Search",
		url = "http://www.onelook.com/?w=%s",
		filter = search.utf8encode,
	},
	LEO = {
		name = "LEO English-German Dictionary",
		url = "http://dict.leo.org/ende?lp=ende&lang=en&searchLoc=0&cmpType=relaxed&sectHdr=on&spellToler=on&relink=on&search=%s",
		filter = search.utf8encode,
	},
	IMDB = {
		name = "Internet Movie Database (English)",
		url = "http://www.imdb.com/find?q=%s",
		filter = search.utf8encode,
	},
	KM = {
		name = "K-Meleon Forums",
		url = "http://kmeleon.sourceforge.net/forum/search.php?f=3&globalsearch=1&match=1&date=0&fldsubject=1&fldbody=1&search=%s",
		filter = search.encode,
	},
}

search.eng = getpref(TYPE_STRING, search.se)
if search.eng == "" then
	search.eng = "Google"
end

local function selectEngine()
	local names, tbl
	
	tbl, names = {}, {}
	table.foreach(search.engines, function (item)
		if search.engines[item].name then
			table.insert(tbl, item)
			table.insert(names, search.engines[item].name)
		end
	end)
	i = choose(search.mes.select_prompt, search.mes.select_caption,
	           unpack(names))
	if i then
		return tbl[i]
	end
end

local function openSearch(engine, word, how)
	local filter, url
	
	filter = search.engines[engine].filter
	if filter and type(filter) == "function" then
		word = filter(word)
	end
	url = string.format(search.engines[engine].url, word)
	if not how then
		how = getpref(TYPE_INT, search.so)
	end
	if how == 1 then
		layer.openNew(url)
	elseif how == 2 then
		layer.openBack(url)
	else 
		NavigateTo(url, OPEN_NORMAL)
	end
end
search.openSearch = openSearch

local function promptSearch(engine, default)
	local tip = ""
	local title
	
	if engine == "URL" then
		tip = search.mes.enter_url
		title = search.mes.go_url
	else
		tip = search.engines[engine].name
		if not tip or tip == "" then
			tip = engine
		end
		tip = string.format(search.mes.prompt_prompt, tip)
		title = search.mes.prompt_caption
	end
	return prompt(tip, title, default)
end
search.promptSearch = promptSearch

local function sync(set)
	if set then
		if search.eng then
			SetCheck("luamacros(search.Search(" .. search.eng .. ")", false)
		end
		search.eng = set
		setpref(TYPE_STRING, search.se, set)
		SetCheck("luamacros(search.Search(" .. set .. ")", true)
	end
end
search.engineSync = sync

local function defaultEngine(engine)
	if engine == "" or not search.engines[engine] then
		engine = search.eng
	end
	if engine == "" then
		engine = "Google"
	end
	if not search.engines[engine].url then
		return nil
	end
	return engine
end
search.defaultEngine = defaultEngine

function search.SetOpen(set)
	local c
	if set == "0" or set == "1" or set == "2" then
		c = getpref(TYPE_INT, search.so)
		if c == 0 or c == 1 or c == 2 then
			SetCheck("luamacros(search.SetOpen(" .. c .. ")", false)
		end
		setpref(TYPE_INT, search.so, set)
	end
	c = getpref(TYPE_INT, search.so)
	if c == 0 or c == 1 or c == 2 then
		SetCheck("luamacros(search.SetOpen(" .. c .. ")", true)
	end
end

function search.SetSearch(set)
	local names, tbl, i
	
	if not set or set == "" then
		set = selectEngine()
	end
	if set and search.engines[set] then
		sync(set)
	end
end

function search.Search(engine)
	local word, title, url
	
	engine = defaultEngine(engine)
	if not engine then
		return
	end
	word = GetGlobalVar(TYPE_STRING, "SelectedText")
	if word and word ~= "" then
		word = promptSearch(engine, word)
		if word and word ~= "" then
			openSearch(engine, word)
		end
		return
	end
	word = GetGlobalVar(TYPE_STRING, "URLBAR")
	url = GetGlobalVar(TYPE_STRING, "URL")
	if word and word ~= "" and word ~= url then
		SetGlobalVar(TYPE_STRING, "URLBAR", url)
		openSearch(engine, word)
		return
	end
	word = promptSearch(engine)
	if word and word ~= "" then
		openSearch(engine, word)
	end
end

function search.SelectSearch()
	local engine
	engine = selectEngine()
	if engine then
		search.Search(engine)
	end
end

if hook then
	hook.add(StartupHook, function ()
		sync(search.eng)
		search.SetOpen()
	end)
end
