--[[

Menu setting module

* about

This module allows user to customize menu with add on library.

* example

do
    local options = {
        {},
        {"Set Default Engine", "luamacros(search.SetSearch)"},
        {},
        {"Open results in current page", "luamacros(search.SetOpen(0))"},
        {"Open results in new page", "luamacros(search.SetOpen(1))"},
        {"Open results in background", "luamacros(search.SetOpen(2))"},
    }
    -- scan registred search engines and append menu
    local k, v
    for k, v in pairs(search.engines) do
        table.insert(options, 1, {
            v.name or k, "luamacros(search.Search(" .. k .. "))",
        })
    end
    -- add menu
    setmenu.define("Main", -3, "&Lua", {
        {"Copy As Link", "luamacros(CopyAsLink)"},
        {"Search", "luamacros(search.Search)"},
        {"Search Option", options},
        {"Proxy", {
            {"No Proxy", "luamacros(proxy.change(no))"},
            {"Default Proxy", "luamacros(proxy.change(gateway))"},
            {"Local Proxy", "luamacros(proxy.change(local))"},
        }},
        {"Page Links", "luamacros(ViewLinks)"},
        {"List Links", "luamacros(ListLinks)"},
    })
end


]]

if not KMEL_PLUGIN_VER or KMEL_PLUGIN_VER < 514 then
	return
end

require "hook"

setmenu = {}

local function append(name, pos, label, item)
	if not item then
		km.SetMenu(name, MENU_SEPARATOR, pos)
		return
	end
	local t = type(item)
	if type(label) ~= "string" then
		return
	end
	if t == "number" or t == "string" then
		km.SetMenu(name, MENU_COMMAND, label, item)
	elseif t == "table" then
		km.SetMenu(name, MENU_POPUP, label)
		local k, v
		for k, v in ipairs(item) do
			append(label, k, v[1], v[2])
		end
	end
end

function setmenu.append(name, pos, label, item)
	append(name, pos, label, item)
	km.RebuildMenu(name)
end

--
-- defime menu
--

local defined_menu = {}

function setmenu.define(name, pos, label, item)
	table.insert(defined_menu, {name, pos, label, item})
end

if hook then
	hook.add(SetupHook, function ()
		i, j  = next(defined_menu, nil)
		while i do
			setmenu.append(j[1], j[2], j[3], j[4])
			i, j = next(defined_menu, i)
		end
		-- purge
		defined_menu = nil
	end)
end
